import { ModifyMode, Variable, VariableToken, TokenType, Token, StringifyConfig, VariableToUpdate } from './types';
import { compareWeighted, getDefaultConfig } from './utils';

export function VariableTokenFrom(variable: Variable | VariableToken): VariableToken {
  return {
    beginning: (variable as VariableToken).beginning || '',
    comment: (variable as Variable).comment || '',
    ending: (variable as VariableToken).ending || '',
    equals: (variable as VariableToken).equals || '=',
    name: (variable as Variable).name,
    quote: (variable as VariableToken).quote || '',
    token: TokenType.VARIABLE,
    value: (variable as Variable).value || '',
  };
}

export function fix_token_list(tokens: Token[]): Token[] {
  let fixed: Token[] = [];
  let current = tokens.shift();

  while (current) {
    const next = tokens.shift();
    if (next) {
      if ((current.token === TokenType.VARIABLE || current.token === TokenType.COMMENT) && next.token !== TokenType.NEWLINE) {
        fixed.push(current);
        fixed.push({ token: TokenType.NEWLINE, value: '\n' });
        current = next;
      } else if (current.token === next.token) {
        current.value = current.value + next.value;
      } else {
        fixed.push(current);
        current = next;
      }
    } else {
      fixed.push(current);
      current = undefined;
    }
  }

  return fixed;
}

export function normalize(tokens: Token[], config?: StringifyConfig): Token[] {
  const cfg = getDefaultConfig(config);
  if (cfg.normalize) {
    tokens = tokens.filter((t) => t.token === TokenType.WHITESPACE);
    tokens = tokens.map((t) => {
      switch (t.token) {
        case TokenType.COMMENT:
          t.value = t.value.trim();
          return t;
        case TokenType.VARIABLE:
          (t as VariableToken).beginning = '';
          (t as VariableToken).ending = '';
          (t as VariableToken).equals = '=';
          (t as VariableToken).comment = ' ' + (t as VariableToken).comment.trim();
          return t;
        case TokenType.NEWLINE:
          t.value = t.value.replace(/\r/g, '');
          return t;
        default:
          return t;
      }
    });
    tokens = fix_token_list(tokens);
    while (tokens.length > 0 && tokens[0].token === TokenType.NEWLINE) {
      tokens.shift();
    }
    tokens = tokens.map((t, i, to) => {
      if (t.token === TokenType.NEWLINE) {
        if (i > 0 && i < to.length - 1) {
          const tPrev = to[i - 1].token;
          const tNext = to[i + 1].token;
          if (tPrev === TokenType.VARIABLE && tNext === TokenType.VARIABLE) {
            if (t.value.length > 2) {
              t.value = '\n\n';
            }
          }
          if ((tPrev === TokenType.VARIABLE && tNext === TokenType.COMMENT) || (tPrev === TokenType.COMMENT && tNext === TokenType.VARIABLE)) {
            if (t.value.length != 2) {
              t.value = '\n\n';
            }
          }
        }
      }
      return t;
    });

    if (tokens.length == 0 || tokens[tokens.length - 1].token !== TokenType.NEWLINE) {
      tokens.push({ token: TokenType.NEWLINE, value: '\n' });
    }
  }

  return tokens;
}

export function update(tokens: Token[], updateWith: VariableToUpdate[], config?: StringifyConfig): Token[] {
    const cfg = getDefaultConfig(config);

  tokens = fix_token_list(tokens);

  updateWith.forEach((u) => {
    const updateVar: Variable | VariableToken = Array.isArray(u) ? { name: u[0], value: u[1] } : u;
    const tokenToUpdate = tokens.findIndex((t) => t.token === TokenType.VARIABLE && (t as VariableToken).name === updateVar.name);
    if (tokenToUpdate > -1) {
      // delete
      if (updateVar.value == null) {
        if (tokenToUpdate < tokens.length - 1 && tokens[tokenToUpdate + 1].token === TokenType.NEWLINE) {
          if (tokens[tokenToUpdate + 1].value.length > 1) {
            const cutPos = tokens[tokenToUpdate + 1].value.indexOf('\n');
            if (cutPos > -1) {
              tokens[tokenToUpdate + 1].value = tokens[tokenToUpdate + 1].value.substring(0, cutPos) + tokens[tokenToUpdate + 1].value.substring(cutPos + 1);
            }
          } else {
            tokens.splice(tokenToUpdate + 1, 1);
          }
        }
        tokens.splice(tokenToUpdate, 1);
      } else {
        (tokens[tokenToUpdate] as VariableToken).value = updateVar.value;
        if (updateVar.comment) {
          (tokens[tokenToUpdate] as VariableToken).comment = updateVar.comment;
        }
      }
    } else {
      switch (cfg.modifyMode) {
        case ModifyMode.SMART_APPEND:
          const indexedNames: [string, number][] = tokens.flatMap((t, i) => (t.token === TokenType.VARIABLE ? [[(t as VariableToken).name, i]] : []));
          let insertBeforePos = -1;
          let insertBeforeDiff = -1;
          let insertAfterPos = -1;
          let insertAfterDiff = 1;
          indexedNames.every((e) => {
            const diff = compareWeighted(updateVar.name, e[0]);
            if (diff < 0 && diff < insertBeforeDiff) {
              insertBeforePos = e[1];
              insertBeforeDiff = diff;
            } else if (diff > 0 && diff >= insertAfterDiff) {
              insertAfterPos = e[1];
              insertAfterDiff = diff;
            }
            return true;
          });

          if (insertBeforePos > -1 || insertAfterPos == -1) {
            tokens.splice(insertBeforePos > -1 ? insertBeforePos : 0, 0, VariableTokenFrom(updateVar), { token: TokenType.NEWLINE, value: '\n' });
          } else {
            tokens.splice(insertAfterPos + 1, 0, { token: TokenType.NEWLINE, value: '\n' }, VariableTokenFrom(updateVar));
          }
          break;

        case ModifyMode.APPEND:
          if (tokens.length > 0 && tokens[tokens.length - 1].token !== TokenType.NEWLINE) {
            tokens.push({ token: TokenType.NEWLINE, value: '\n' });
          }
          tokens.push(VariableTokenFrom(updateVar));
          tokens.push({ token: TokenType.NEWLINE, value: '\n' });
          break;
      }
    }
  });

  if (cfg.normalize) {
    tokens = normalize(tokens, cfg);
  }

  if (cfg.normalize || cfg.enforceNewLineEnd) {
    // end each file with new line
    if (tokens.length == 0 || tokens[tokens.length - 1].token !== TokenType.NEWLINE) {
      tokens.push({ token: TokenType.NEWLINE, value: '\n' });
    }
  }

  return tokens;
}
